"""Grid layout item."""

# pylint: disable=duplicate-code

from __future__ import annotations

import logging

import wx

from aio_wx_widgets.core.base_widget import CallableItem
from aio_wx_widgets.core.sizers import SizerMixin

_LOGGER = logging.getLogger(__name__)

__all__ = ["Grid", "VERTICAL", "HORIZONTAL"]

VERTICAL = wx.VERTICAL
HORIZONTAL = wx.HORIZONTAL


class Grid(SizerMixin, CallableItem):
    """A grid component.

    Basically a boxsizer with some defaults.
    """

    default_sizer_margin = 0

    HOR = HORIZONTAL
    VER = VERTICAL

    def __init__(self, orientation: int = HORIZONTAL):
        """Init."""
        super().__init__()
        self._sizer_ = wx.BoxSizer(orient=orientation)
        self._parent = None

    @property
    def ui_item(self):
        """Return UI item."""
        return self._sizer_

    @property
    def _sizer(self):
        return self._sizer_

    def init(self, parent: wx.Window):
        self._parent = parent

    def __call__(self, parent: wx.Window):
        # setting a specific parent.
        self.init(parent)
        return self

    def __enter__(self):
        """Enter context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        """Exit context manager."""



"""A fixed grid.

when resizing windows this layout seems to be most reliable.
"""

from __future__ import annotations

import logging

import wx

from aio_wx_widgets.core.base_widget import CallableItem
from aio_wx_widgets.core.sizers import ItemType

_LOGGER = logging.getLogger(__name__)


class FixedGrid(CallableItem):
    """A fixed grid implementation.

    When resizing windows this grid behaves best.
    """

    def __init__(self):
        self._sizer_ = wx.GridSizer(1, 0, (0, 0))
        self._parent = None

    @property
    def ui_item(self):
        return self._sizer_

    @property
    def _sizer(self) -> wx.GridSizer:
        return self._sizer_

    def init(self, parent: wx.Window):
        self._parent = parent

    def __call__(self, parent: wx.Window):
        self.init(parent)
        return self

    def __enter__(self) -> FixedGrid:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # type:ignore
        """Exit the context manager."""

    def _check_columns(self):
        used_cols: int = self._sizer_.GetEffectiveColsCount()
        _LOGGER.debug("Used cols %s", used_cols)
        total_cols: int = self._sizer_.GetCols()
        if total_cols == used_cols:
            self._sizer_.SetCols(used_cols + 1)

    def add(self, item: ItemType, create: bool = True) -> ItemType:
        """Add an item to this grid."""
        self._check_columns()
        if create:
            assert isinstance(item, CallableItem)
            item.init(self._parent)

        self._sizer_.Add(item.ui_item, 1, wx.EXPAND)  # type:ignore
        return item

    def add_spacer(self):
        """Add an empty space to a column."""
        self._check_columns()
        self._sizer_.Add(wx.Panel(self._parent), 1, flag=wx.EXPAND)  # type: ignore


class DynamicGrid(CallableItem):
    """A dynamic grid. Allows for column spanning etc."""

    def __init__(self):
        self._sizer_ = wx.GridBagSizer()
        self._sizer_.SetFlexibleDirection(wx.VERTICAL)
        self._sizer_.SetNonFlexibleGrowMode(wx.FLEX_GROWMODE_ALL)
        self._parent = None

    @property
    def ui_item(self):
        return self._sizer_

    @property
    def _sizer(self):
        return self._sizer_

    def init(self, parent: wx.Window):
        self._parent = parent
        # self._parent.Bind(wx.EVT_SIZE,self._on_size)

    def _on_size(self, evt: wx.SizeEvent):
        evt.Skip()

        _LOGGER.debug(
            "Laying out dynamic grid. %s, sizer width %s, parent sizxe",
            evt.Size,
            self._parent.Size,
        )
        # wx.CallAfter(self.ui_item.PostSizeEvent())

    def __call__(self, parent: wx.Window):
        self.init(parent)
        return self

    def __enter__(self) -> DynamicGrid:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        """Exit the context manager."""

    def add(
        self, item: ItemType, col: int, span: int = 1, create: bool = True
    ) -> ItemType:
        """Add an item to this grid."""

        if create:
            assert isinstance(item, CallableItem)
            item.init(self._parent)

        self._sizer_.Add(item.ui_item, (0, col), (1, span), flag=wx.EXPAND)  # type: ignore
        self._set_col_sizes_equal()
        return item

    def _set_col_sizes_equal(self):
        for idx in range(self._sizer_.Cols):
            if not self._sizer_.IsColGrowable(idx):  # type: ignore
                self._sizer_.AddGrowableCol(idx)  # type: ignore

    def add_spacer(self, col, span=1):
        """Add an empty space to a column and/or span."""
        self._sizer_.Add(wx.Panel(self._parent), (0, col), (1, span), flag=wx.EXPAND)
        self._set_col_sizes_equal()



import logging

from demo.controller.text_wrapping_controller import TextWrappingController

from aio_wx_widgets.panels.panel import SimplePanel
from aio_wx_widgets.widgets import LabelledItem, Text
from aio_wx_widgets.widgets.experimental.bullet_text import BulletText
from aio_wx_widgets.widgets.experimental.fixed_grid import DynamicGrid, FixedGrid

_LOGGER = logging.getLogger(__name__)

LONG_TEXT = 10 * "very very long text that should just wrap.."


class TextWrappingView(SimplePanel[TextWrappingController]):
    """A text wrapping view test."""

    def populate(self):
        with self.add(DynamicGrid()) as grd:
            grd.add(Text(text=LONG_TEXT, wrap=True), col=1)
            grd.add(Text(text=LONG_TEXT, wrap=True), col=2)
        with self.add(FixedGrid()) as grd:
            grd.add_spacer()
            grd.add(Text(text=LONG_TEXT, wrap=True))
            grd.add_spacer()
        self.add(
            LabelledItem(
                label_text="This is the label", item=Text("Just text"), item_weight=3
            )
        )
        with self.add(FixedGrid()) as grd:
            grd.add_spacer()
            grd.add(BulletText(text=LONG_TEXT, wrap=True))
            grd.add_spacer()



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import wx

from aio_wx_widgets.colors import BLACK, RED
from aio_wx_widgets.containers.grid import Grid
from aio_wx_widgets.containers.group import Section
from aio_wx_widgets.panels.panel import SimplePanel
from aio_wx_widgets.widgets.button import AioButton
from aio_wx_widgets.widgets.text import Text

if TYPE_CHECKING:
    from demo.controller.controller_one import ControllerOne

_LOGGER = logging.getLogger(__name__)

IMAGE_FOLDER = (Path(__file__).parent) / "image"


def _bool_to_color(value) -> int:
    """Convert a boolean to a color value.

    Used in a binding.
    """
    if value is True:
        return RED
    return BLACK


class ViewOne(SimplePanel["ControllerOne"]):
    def __init__(self, parent, controller: "ControllerOne"):
        super().__init__(parent, controller, scrollable=True)

    def populate(self):
        """Populate this view."""

        # self.add(BulletText("this is stupid text."))
        # with self.add(FixedGrid()) as grd:
        #     grd.add(Text("This is text"))
        #     grd.add(Text("this is other text"))
        #     grd.add(Text("This is text"))
        #
        # with self.add(FixedGrid(), weight=1, margin=0) as fixed:
        #     fixed.add(Text("Col1"))
        #     fixed.add(Text("col2"))
        #     fixed.add(Image(IMAGE_FOLDER / "phoenix_main.png"))
        #
        # with self.add(Group("label")) as group:
        #     group.add(Text(binding=self.bind("int_val"), wrap=True), margin=20)
        #     # with group.add(Group("other group")) as grp:
        #     #     grp.add(Text(text="more text"))
        #     # with group.add(Section("section inside group")) as sec:
        #     #     sec.add(Text(text="section inside group"))
        #     # with group.add(Grid(orientation=Grid.VER),weight=1) as grd:
        #     #     grd.add(Text("This is long text", min_width=50))
        #     #     grd.add(Text("short", min_width=50))
        #
        # self.add(
        #     LabelledItem(
        #         "Label text",
        #         Entry(binding=self.bind("float_val"), validator=float_validator),
        #     )
        # )
        #
        # with self.add(Section("Float validators")) as sc:
        #     sc.add(
        #         LabelledItem(
        #             "Enter a float value",
        #             Entry(binding=self.bind("float_val"), validator=float_validator,),
        #             item_weight=2,
        #         )
        #     )
        #     with sc.add(Grid()) as grd:
        #         # the binding binds to an attribute defined in the controller
        #         # the weight determines how much space a specific item should consume
        #         # with respect to the other members of the container.
        #
        #         grd.add(
        #             Entry(binding=self.bind("float_val"), validator=float_validator,),
        #             weight=4,
        #         )
        #         grd.add(
        #             Text(binding=self.bind("float_val")),
        #             weight=4,
        #             ver_align=VerAlign.center,
        #         )
        # with self.add(Section("Int validators")) as sc:
        #     sc.add(Text("Only integer values are allowed."))
        #     with sc.add(Grid()) as grd:
        #         grd.add(
        #             Entry(binding=self.bind("int_val"), validator=int_validator,),
        #             weight=3,
        #             margin=0,
        #         )
        #         grd.add(
        #             Entry(binding=self.bind("int_val"), validator=int_validator,),
        #             weight=1,
        #             margin=0,
        #         )
        #         grd.add(Text(binding=self.bind("int_val")), weight=1, margin=0)
        #
        # with self.add(Section("Bindings")) as sec:
        #     with sec.add(Grid()) as grd:
        #         grd.add(
        #             Text(
        #                 "Use a converter with a binding to bind different types "
        #                 "of properties. Bind a color to a boolean using a converter.",
        #                 wrap=True,
        #             ),
        #             weight=1,
        #         )
        #         with grd.add(Grid(), weight=1) as grd1:
        #             grd1.add(CheckBox("RED", self.bind("color_red")))
        #             grd1.add(
        #                 Text(
        #                     text="This is text with a color binding",
        #                     color=self.bind("color_red", converter=_bool_to_color),
        #                 )
        #             )
        #     with sec.add(Grid()) as grd:
        #         grd.add(
        #             Text("Bind to the enabled property of a widget."),
        #             weight=1,
        #             ver_align=VerAlign.center,
        #         )
        #         with grd.add(Grid(), weight=1) as items:
        #             btn = items.add(AioButton("toggle", self._toggle))
        #             print(btn.ui_item.Parent)
        #             items.add(
        #                 AioButton("toggle", self._toggle, enabled=self.bind("ready"))
        #             )
        #             items.add(
        #                 Entry(self.bind("a_string_value"), enabled=self.bind("ready"))
        #             )
        # with self.add(Grid()) as grd:
        #     grd.add(Entry(binding=Binding(self._controller, "a_string_value")))
        #     grd.add(Entry(binding=Binding(self._controller, "a_string_value")))
        #
        # with self.add(Grid()) as grd:
        #     grd.add(AioButton("button one", self._set_value), weight=6)
        #     grd.add(AioButton("button two,self", self._set_value), weight=6)
        #
        # with self.add(Section("Layout")) as grp:
        #     with grp.add(Grid()) as grd:
        #         grd.add(
        #             Text(text="Widget, right & center alignment"),
        #             weight=3,
        #             hor_align=HorAlign.right,
        #             ver_align=VerAlign.center,
        #         )
        #         with grd.add(Grid(VERTICAL), weight=6, margin=0) as vert_grid:
        #             vert_grid.add(
        #                 AioButton("Left aligned button", self._set_value),
        #                 hor_align=HorAlign.left,
        #                 margin=4,
        #             )
        #             vert_grid.add(
        #                 AioButton(
        #                     "No Aligned widget (button in this case)", self._set_value
        #                 )
        #             )
        #             vert_grid.add(
        #                 Text(text="Center aligned widget with a large margin."),
        #                 margin=(10, 10, 30, 5),
        #                 hor_align=HorAlign.center,
        #             )
        #             vert_grid.add(
        #                 Text(
        #                     text="Text widget with centered text",
        #                     hor_align=Text.HOR_CENTER,
        #                 )
        #             )
        #             vert_grid.add(
        #                 AioButton(
        #                     "right aligned button.",
        #                     self._set_value,
        #                     enabled=self.bind("a_checkbox_value"),
        #                 ),
        #                 weight=2,
        #                 hor_align=HorAlign.right,
        #             )
        with self.add(Section("Min width")) as grp:
            with grp.add(Grid()) as grd:
                grd.add(Text("This is a text exceeding 80 min size"))
                grd.add(Text("Other text."))

            with grp.add(Grid()) as grd:
                grd.add(Text("text < 80", min_width=80))
                grd.add(Text("Other text."))

        txt = wx.StaticText(
            parent=self.ui_item, label="A long label exceeding minwidth"
        )

        with grp.add(Grid()) as grd:
            grd.add(txt, create=False)
            grd.add(Text("Other text."))
        txt.SetMinSize(wx.Size(80, -1))

        # self.add(CheckBox("A checkbox", binding=self.bind("a_checkbox_value")))
        # self.add(
        #     CheckBox("The same checkbox value", binding=self.bind("a_checkbox_value"))
        # )
        self.add(AioButton("open other window", self._on_open_second_window))
        self.add(AioButton("open third window", self._on_open_third_window))
        self.add(
            AioButton("open text wrapping window", self._on_open_text_wrapping_window)
        )
        self.add(AioButton("Open image view", self._on_open_image_view))

    def _toggle(self, evt: wx.Event):
        self.controller.ready = not self.controller.ready

    async def _on_open_second_window(self, evt: wx.Event):
        await self._controller.open_other_window()

    async def _on_open_third_window(self, evt: wx.Event):
        await self._controller.open_third_window()

    async def _on_open(self, evt: wx.Event):
        await self._controller.open_other_window()

    async def _on_open_text_wrapping_window(self, evt: wx.Event):
        await self._controller.open_text_wrapping_window()

    async def _on_open_image_view(self, evt: wx.Event):
        await self._controller.open_image_window()

    async def _set_value(self, evt: wx.Event):
        await self._controller.set_value()



import logging

from demo.controller.image_view_controller import ImageViewController
from demo.model.demo_model import DemoModel
from demo.views.image_view import ImageView

from aio_wx_widgets.frame import DefaultFrame

_LOGGER = logging.getLogger(__name__)


class ImageViewWindow(DefaultFrame):
    def __init__(self):
        super().__init__("Text wrapping window")
        model = DemoModel()
        controller = ImageViewController(model)
        self.view = ImageView(self, controller)
        self.view.populate()



"""All type annotations go here."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Union

try:
    import Protocol  # type: ignore
except ImportError:
    from typing_extensions import Protocol


class Choice(Protocol):
    """Typing protocol to be used for static type checking."""

    label: str
    value: Any


if TYPE_CHECKING:
    from asyncio import AbstractEventLoop

    import wx

    from aio_wx_widgets.controller import BaseController
    from aio_wx_widgets.core.base_widget import BaseWidget, CallableItem
    from aio_wx_widgets.core.binding import Binding
    from aio_wx_widgets.core.sizers import PanelMixin

    assert PanelMixin

    assert BaseController
    assert AbstractEventLoop
    assert Binding
    assert wx
    assert BaseWidget

    Widget = Union[BaseWidget, wx.Window, wx.BoxSizer, CallableItem]
    ConverterType = Callable[[Any], Union[float, int, str, bool]]



"""Experimental bullet text"""

from aio_wx_widgets.containers import Grid
from aio_wx_widgets.widgets import Text


class BulletText(Grid):
    """Text with a bullet."""

    def __init__(self, text: str, font_scale=1, bullet: str = "-", wrap=True):
        self._text = str(text)
        self._font_scale = font_scale
        self._bullet = bullet
        self._wrap = wrap

        super().__init__()

    def init(self, parent, *args, **kwargs):
        super().init(parent)
        self.add(Text(self._bullet, font_size=self._font_scale))
        self.add(Text(self._text, font_size=self._font_scale, wrap=True), weight=1)
        return self



"""Image widget."""

import logging
from pathlib import Path

import wx

from aio_wx_widgets.colors import RED
from aio_wx_widgets.const import is_debugging
from aio_wx_widgets.core.base_widget import BaseWidget

_LOGGER = logging.getLogger(__name__)

__all__ = ["Image"]


def _get_ratio(image: wx.Image):
    size = image.GetSize()
    width_height = size[0] / size[1]
    return width_height


class _SizeableImage(wx.StaticBitmap):
    """A static bitmap child.

    The DoGetBestClientSize override method ensures that the proper image size
    is set once its parent(s) start resizing.
    """

    def __init__(self, *args, **kwargs):
        self._image_ratio = kwargs.pop("ratio")
        self._image = kwargs.pop("image")
        self._min_width = 10
        self._prev_image_size = (-1, -1)

        super().__init__(*args, **kwargs)
        if is_debugging():
            self.SetBackgroundColour(RED)

    def _set_image(self, size):
        """Set the image."""
        if size[0] <= 0 or size[1] <= 0:
            _LOGGER.debug("Image size has 0 value. Skipping")
            return
        if self._prev_image_size == size:
            _LOGGER.debug("Skipping image resize")
            return

        self._prev_image_size = size
        _LOGGER.debug("Setting image size to : %s", size)
        image = self._image.Scale(size[0], size[1], quality=wx.IMAGE_QUALITY_BICUBIC)
        bitmap = wx.Bitmap(image)

        self.SetBitmap(bitmap)

    # pylint: disable=invalid-name
    def DoGetBestClientSize(self):
        """Return best image size when parent sizer is re-arranging children."""
        item_size = self.GetSize()
        _LOGGER.debug("Image size %s", item_size)

        min_x = item_size[0]

        optimal_size = (-1, int(min_x / self._image_ratio))
        image_size = (min_x, optimal_size[1])
        # if not image_size == self._prev_image_size:
        self._set_image(image_size)

        return optimal_size


class Image(BaseWidget):
    """Autoscaling image widget.

    The image will resize once it is put inside a boxsizer (or aio-wx-widget grid).
    """

    def __init__(self, image: Path, min_width=10, max_width=None):
        """Init.

        Args:
            image: The image to be displayed.
            min_width: The minimal allowed image size.
            max_width: Max allowed image width.
        """
        self._image = wx.Image(str(image), wx.BITMAP_TYPE_PNG)
        self._image_ratio = _get_ratio(self._image)
        super().__init__(
            _SizeableImage(ratio=self._image_ratio, image=self._image),
            min_width=min_width,
            value_binding=None,
        )

    def init(self, parent):
        self.ui_item.Create(parent)

    def __call__(self, parent):
        self.init(parent)
        return self



from __future__ import annotations

import logging
from random import randint

from demo.model.demo_model import DemoModel

from aio_wx_widgets.controller import BaseController

_LOGGER = logging.getLogger(__name__)


class ControllerThree(BaseController[DemoModel]):
    def __init__(self, model: DemoModel):
        self.value_1: int = 0
        self.a_string_value = "A certain string"
        super().__init__(model)

        # bindable properties.
        self.ready = False
        self.bound_text = None

    async def set_value(self):
        val = randint(1, 100)
        _LOGGER.debug("Setting bound property to %s", val)
        self.value_1 = val

    @property
    def demo(self) -> DemoModel:
        return self._model
